import pytest
from unittest.mock import patch
from collections import Counter
from src.key_mwe.ngram_dict_tokeniser import NgramDictTokeniser


@pytest.fixture
def sample_blacklist() -> list[str]:
    return ['a', 'an', 'the']


@pytest.fixture
def sample_mwe_range() -> list[int]:
    return [2, 3]


@pytest.fixture
def ngram_dict_tokeniser(sample_mwe_range: list[int], sample_blacklist: list[str]) -> NgramDictTokeniser:
    return NgramDictTokeniser(sample_mwe_range, sample_blacklist)


def test_update_counts(ngram_dict_tokeniser: NgramDictTokeniser):
    sentence = ["This", "is", "a", "test", "sentence"]
    ngram_dict_tokeniser.update_counts(sentence)
    expected = {
        1: Counter({"This": 1, "is": 1, "test": 1, "sentence": 1}),  # 'a' is in blacklist
        2: Counter({"This is": 1, "test sentence": 1}),  # 'a' and 'is a' are affected by blacklist
        3: Counter({"This is test": 1})  # Only one valid 3-gram without blacklist words
    }
    assert ngram_dict_tokeniser.get_ngrams() == expected


@pytest.mark.parametrize("corpus, expected", [
    (["This", "is", "a", "test"], {1: Counter({'This': 1, 'is': 1, 'a': 1, 'test': 1}), 
                                   2: Counter({"This is": 1}),
                                   3: Counter({"is a test": 1})}),
    ([], {1: Counter(), 2: Counter(), 3: Counter()}),  # Testing with empty input
    (["This", "is", "another", "test"], {1: Counter({"This": 1, "is": 1, "another": 1, "test": 1}),
                                         2: Counter({'This is': 1, 'is another': 1, 'another test': 1}),
                                         3: Counter({'This is another': 1, 'is another test': 1})}),
    (["With", "two", "sentences"], {1: Counter({"With": 1, "two": 1, "sentences": 1}),
                                    2: Counter({"With two": 1, "two sentences": 1}),
                                    3: Counter({"With two sentences": 1})}),
    (["hello", "zuluzulu", "world", "zuluzulu", "zuluzulu", "cómo", "están", "zuluzulu", "numbers", "zuluzulu", "1234", "zuluzulu"],
     {1: Counter({"hello": 1, "world": 1, "cómo": 1, "están": 1, "numbers": 1, "1234": 1}),
        2: Counter({"cómo están": 1}),
        3: Counter({})})
])
def test_update_counts(ngram_dict_tokeniser: NgramDictTokeniser, corpus: list[list[str]], expected: dict[int, Counter]):
    ngram_dict_tokeniser.update_counts(corpus)
    assert ngram_dict_tokeniser.get_ngrams() == expected


def test_get_ngram_counts(ngram_dict_tokeniser: NgramDictTokeniser):
    # Directly setting the ngrams attribute to test get_ngram_counts
    ngram_dict_tokeniser.ngrams = {1: Counter({'test': 3, 'dummy': 1}), 2: Counter({'test dummy': 2})}
    assert ngram_dict_tokeniser.get_ngram_counts(1) == Counter({'test': 3, 'dummy': 1})
    assert ngram_dict_tokeniser.get_ngram_counts(2) == Counter({'test dummy': 2})
    with pytest.raises(KeyError):
        ngram_dict_tokeniser.get_ngram_counts(3)  # Test for non-existent key



import math
import logging


from collections import Counter


class NpmiEstimator:
    
    LOG_MIN_VALUE = 1e-10
    LOG_MIN = math.log(LOG_MIN_VALUE)
    

    def __init__(self, ngrams: dict[int, Counter]) -> None:
        self.ngrams: dict[int, Counter] = {n: counter for n, counter in ngrams.items() if counter}
        self.ngram_probs: dict[int, dict[str, float]] = {n: {} for n in self.ngrams.keys()}
        self.npmi_values: dict[int, dict[str, float]] = {n: {} for n in self.ngrams.keys()}
        self.total_counts: dict[int, int] = {n: sum(self.ngrams[n].values()) for n in self.ngrams.keys()}


    def estimate_within_corpus_npmi(self, adjusted: bool = True) -> None:
        for n in self.ngrams.keys():
            for ngram in self.ngrams[n].keys():
                self.estimate_ngram_npmi(n, ngram, adjusted)


    def estimate_ngram_npmi(self, n: int, ngram: str, adjusted: bool) -> None:
        tokens: list[str] = ngram.split()     
        if len(tokens) != n:
            logging.warning(f'Unexpected number of tokens for n-gram: {ngram}')
            return None
        # prob of ngram in corpus
        p_ngram: float = self.get_ngram_prob(n, ngram)
        # If only one ngram in an ngram size, assume an npmi of "neutral":
        if p_ngram == 1:
            self.npmi_values[n][ngram] = 0
            return
        # prob of tokens in corpus
        if n > 1:
            p_tokens: list[float] = [self.get_ngram_prob(1, token) for token in tokens]
            log_p_ngram: float = self.safe_log(p_ngram)
            log_p_tokens: float = [self.safe_log(p) for p in p_tokens]
            # Calculate PMI
            log_prod_p_tokens: float = sum(log_p_tokens)  # log(a*b) = log(a) + log(b)
            pmi: float = log_p_ngram - log_prod_p_tokens  # log(a/b) = log(a) - log(b)
            # Normalizing PMI
            npmi: float = pmi / -log_p_ngram
            # Alternative approach to PMI calculation:
            # prod_p_tokens: float = math.prod(p_tokens)
            # pmi: float = math.log(p_ngram / prod_p_tokens)
            # npmi: float = pmi / -math.log(p_ngram)
            # Adjusted NPMI Calculation
            if adjusted:
                length_factor: float = math.log(1 + n)
                freq_factor: float = math.log(1 + self.ngrams[n][ngram])
                npmi: float = (npmi / length_factor) * freq_factor
            # Caching the calculated NPMI value
            self.npmi_values[n][ngram] = npmi

    
    def get_ngram_prob(self, n: int, ngram: str) -> float:
        if n in self.ngrams.keys():
            ngrams_n: Counter = self.ngrams[n]
            ngram_probs_n: dict[int, dict[str, float]] = self.ngram_probs[n]
            if ngram not in ngram_probs_n:
                ngram_probs_n[ngram] = ngrams_n[ngram] / self.total_counts[n]
            return ngram_probs_n[ngram]
        else:
            return self.LOG_MIN_VALUE


    def get_npmi_value(self, n: int, ngram: str) -> float | None:
        if n in self.ngrams.keys():
            if ngram not in self.npmi_values[n]:
                self.estimate_ngram_npmi(n, ngram)
            return self.npmi_values[n].get(ngram)
        return None
    

    def get_sorted_unigram_probs(self, top_n: int = None, threshold: float = None, reverse: bool = True) -> list[tuple[str, float]]:
        sorted_unigrams: list[tuple[str, float]] = sorted(self.ngram_probs[1].items(), key=lambda item: item[1], reverse=reverse)
        if threshold:
            sorted_unigrams = [item for item in sorted_unigrams if item[1] > threshold]
        if top_n:
            sorted_unigrams = sorted_unigrams[:top_n]
        return sorted_unigrams


    def get_sorted_npmi_values(self, top_n: int = None, threshold: float = None, reverse: bool = True) -> dict[int, list[tuple[str, float]]]:
        sorted_values = {}
        for n, ngram_dict in self.npmi_values.items():
            sorted_values[n] = sorted(ngram_dict.items(), key=lambda item: item[1], reverse=reverse)
            if threshold:
                sorted_values[n] = [item for item in sorted_values[n] if item[1] > threshold]
            if top_n:
                sorted_values[n] = sorted_values[n][:top_n]
        return sorted_values


    def get_unigrams_and_collocations(
            self, 
            top_n_unigrams: int = None, 
            threshold_probs_unigrams: int = None,
            top_n_collocations: int = None, 
            threshold_npmi_collocations: float = None, 
            reverse: bool = True
        ) -> dict[int, list[tuple[str, float]]]:
        sorted_unigrams: list[tuple[str, float]] = self.get_sorted_unigram_probs(
            top_n=top_n_unigrams, 
            threshold=threshold_probs_unigrams, 
            reverse=reverse
            )
        sorted_collocations: dict[int, list[tuple[str, float]]] = self.get_sorted_npmi_values(
            top_n=top_n_collocations, 
            threshold=threshold_npmi_collocations, 
            reverse=reverse
            )
        if len(sorted_unigrams) > 0:
            sorted_collocations[1] = sorted_unigrams
        return sorted_collocations


    @staticmethod
    def safe_log(value: float) -> float:
        if value < NpmiEstimator.LOG_MIN_VALUE:
            return NpmiEstimator.LOG_MIN
        return math.log(value)



from gensim.models.word2vec import LineSentence
from collections import Counter
from collections.abc import Iterator

from src.key_mwe.config import SEPARATOR_TOKEN
from src.key_mwe.text_preprocessor import Preprocessor


class NgramDictTokeniser:

    def __init__(self, mwe_range: list[int], blacklist: list[str] | set[str]) -> None:
        self.n_range: list[int] = [1] + mwe_range
        self.blacklist: set[str] = self._set_blacklist(blacklist)
        self.ngrams: dict[int, Counter] = {n: Counter() for n in self.n_range}


    def _set_blacklist(self, blacklist: list[str] | set[str]) -> set[str]:
        blacklist: set[str] = set(blacklist) | {SEPARATOR_TOKEN, SEPARATOR_TOKEN.lower()}
        return blacklist


    def tokenise_corpus_from_text_file(self, corpus_file: str) -> None:
        # Assumption: Text file has had text already preprocessed elsewhere.
        for sentence in LineSentence(corpus_file):
            self.update_counts(sentence)


    def tokenise_corpus_from_iterator(self, sentences: Iterator[str]) -> None:
        preprocessor = Preprocessor()
        for sentence in sentences:
            sentence_processed: str = preprocessor.clean_line(sentence)
            self.update_counts(sentence_processed.split())


    def update_counts(self, sentence: list[str]):
        for n in self.n_range:
            if n == 1:
                self.ngrams[n].update([token for token in sentence if token not in [SEPARATOR_TOKEN, SEPARATOR_TOKEN.lower()]])
            else:
                ngrams_sentence = [' '.join(sentence[i: i+n]) for i in range(len(sentence) - n + 1)
                                    if sentence[i] not in self.blacklist and sentence[i+n-1] not in self.blacklist and SEPARATOR_TOKEN not in sentence[i: i+n] and SEPARATOR_TOKEN.lower() not in sentence[i: i+n]]
                self.ngrams[n].update(ngrams_sentence)


    def get_ngrams(self) -> dict[int, Counter]:
        return self.ngrams


    def get_ngram_counts(self, n: int) -> Counter:
        return self.ngrams[n]



import pytest
from collections import Counter
from src.key_mwe.npmi_estimator import NpmiEstimator

@pytest.fixture
def basic_ngrams() -> dict[int, Counter]:
    return {
        1: Counter({'test': 4, 'example': 2}),
        2: Counter({'test example': 2, 'example test': 1})
    }

@pytest.fixture
def complex_ngrams() -> dict[int, Counter]:
    return {
        1: Counter({'test': 5, 'example': 3, 'sample': 2}),
        2: Counter({'test example': 3, 'example test': 2, 'test sample': 1}),
        3: Counter({'test example sample': 1})
    }

@pytest.fixture
def npmi_estimator_basic(basic_ngrams: dict[int, Counter]) -> NpmiEstimator:
    return NpmiEstimator(basic_ngrams)


@pytest.fixture
def npmi_estimator_complex(complex_ngrams: dict[int, Counter]) -> NpmiEstimator:
    return NpmiEstimator(complex_ngrams)


def test_initialization_basic(npmi_estimator_basic: NpmiEstimator) -> None:
    assert isinstance(npmi_estimator_basic, NpmiEstimator)
    assert npmi_estimator_basic.ngrams == {
        1: Counter({'test': 4, 'example': 2}),
        2: Counter({'test example': 2, 'example test': 1})
    }


def test_initialization_complex(npmi_estimator_complex: NpmiEstimator) -> None:
    assert isinstance(npmi_estimator_complex, NpmiEstimator)
    assert npmi_estimator_complex.ngrams == {
        1: Counter({'test': 5, 'example': 3, 'sample': 2}),
        2: Counter({'test example': 3, 'example test': 2, 'test sample': 1}),
        3: Counter({'test example sample': 1})
    }


def test_estimate_ngram_npmi_basic_estimates_npmi_one_ngram_not_unigram(npmi_estimator_basic: NpmiEstimator) -> None:
    npmi_estimator_basic.estimate_ngram_npmi(2, 'test example', False)
    assert 'test example' in npmi_estimator_basic.npmi_values[2]


def test_estimate_within_corpus_npmi_basic_estimates_npmi_for_all_ngrams_not_unigrams(npmi_estimator_basic: NpmiEstimator) -> None:
    npmi_estimator_basic.estimate_within_corpus_npmi()
    assert all(ngram in npmi_estimator_basic.npmi_values[n] for n in npmi_estimator_basic.ngrams if n > 1 for ngram in npmi_estimator_basic.ngrams[n])


def test_get_ngram_prob_basic(npmi_estimator_basic: NpmiEstimator) -> None:
    prob: float = npmi_estimator_basic.get_ngram_prob(1, 'test')
    assert prob == 4/6


def test_get_sorted_npmi_values_basic_adjusted_true(npmi_estimator_basic: NpmiEstimator) -> None:
    npmi_estimator_basic.estimate_within_corpus_npmi(adjusted=True)
    sorted_values: dict[int, list[tuple[str, float]]] = npmi_estimator_basic.get_sorted_npmi_values()
    assert isinstance(sorted_values, dict)
    assert all(isinstance(sorted_values[n], list) for n in sorted_values)
    top_bigram: tuple[str, float] = sorted_values[2][0]
    top_bigram = (top_bigram[0], round(top_bigram[1], 2))
    assert top_bigram == ("test example", round(2.7095112913514545, 2))


def test_get_sorted_npmi_values_basic_adjusted_false(npmi_estimator_basic: NpmiEstimator) -> None:
    npmi_estimator_basic.estimate_within_corpus_npmi(adjusted=False)
    sorted_values: dict[int, list[tuple[str, float]]] = npmi_estimator_basic.get_sorted_npmi_values()
    assert isinstance(sorted_values, dict)
    assert all(isinstance(sorted_values[n], list) for n in sorted_values)
    top_bigram: tuple[str, float] = sorted_values[2][0]
    top_bigram = (top_bigram[0], round(top_bigram[1], 2))
    assert top_bigram == ("test example", round(2.7095112913514545, 2))


def test_unexpected_ngram_length_basic(npmi_estimator_basic: NpmiEstimator) -> None:
    npmi_estimator_basic.estimate_ngram_npmi(2, 'unexpected bigram', False)
    assert 'this is unexpected' not in npmi_estimator_basic.npmi_values[2]


def test_safe_log_edge_cases(npmi_estimator_basic: NpmiEstimator) -> None:
    assert npmi_estimator_basic.safe_log(NpmiEstimator.LOG_MIN_VALUE) == NpmiEstimator.LOG_MIN
    assert npmi_estimator_basic.safe_log(0) == NpmiEstimator.LOG_MIN


def test_get_sorted_unigram_probs_no_params(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    sorted_unigrams = npmi_estimator_complex.get_sorted_unigram_probs()
    assert sorted_unigrams == [('test', 0.5), ('example', 0.3), ('sample', 0.2)], "Unigrams should be sorted by probability in descending order."


def test_get_npmi_value_single_ngram_in_ngram_size_returns_zero(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    npmi_value = npmi_estimator_complex.get_npmi_value(3, 'test example sample')
    assert npmi_value == 0, "NPMI value should be zero for a single ngram in an ngram size."


def test_get_sorted_unigram_probs_with_threshold(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    sorted_unigrams = npmi_estimator_complex.get_sorted_unigram_probs(threshold=0.2)
    assert sorted_unigrams == [('test', 0.5), ('example', 0.3)], "Unigrams should be filtered by a probability greater than the threshold."


def test_get_sorted_unigram_probs_with_top_n(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    sorted_unigrams = npmi_estimator_complex.get_sorted_unigram_probs(top_n=2)
    assert sorted_unigrams == [('test', 0.5), ('example', 0.3)], "Only the top N unigrams should be returned."


def test_get_sorted_unigram_probs_with_reverse(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    sorted_unigrams = npmi_estimator_complex.get_sorted_unigram_probs(reverse=False)
    assert sorted_unigrams == [('sample', 0.2), ('example', 0.3), ('test', 0.5)], "Unigrams should be sorted by probability in ascending order when reverse is False."


def test_get_sorted_npmi_values_no_params(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi(adjusted=True)
    sorted_npmi = npmi_estimator_complex.get_sorted_npmi_values()
    # There should not be npmi values for unigrams:
    assert len(sorted_npmi[1]) == 0
    # Round the NPMI values to 2 decimal places for safe comparison:
    bigram_npmi = sorted_npmi[2]
    bigram_npmi = [(item[0], round(item[1], 2)) for item in bigram_npmi]
    assert bigram_npmi == [('test example', round(2.1918065485787688, 2)), ('example test', round(0.7268330278608419, 2)), ('test sample', round(0.17987631177945834, 2))]
    # Since there is only one trigram, it's npmi was assumed to be zero (neutral):
    trigram_npmi = sorted_npmi[3]
    assert trigram_npmi == [('test example sample', 0)]


def test_get_sorted_npmi_values_with_threshold(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    threshold = 0.5
    sorted_npmi = npmi_estimator_complex.get_sorted_npmi_values(threshold=threshold)
    # We can only test the bigram NPMI values since the trigram NPMI value is zero, and unigrams are not included:
    bigram_npmi = sorted_npmi[2]
    bigram_npmi = [(item[0], round(item[1], 2)) for item in bigram_npmi]
    assert bigram_npmi == [('test example', round(2.1918065485787688, 2)), ('example test', round(0.7268330278608419, 2))]
    # Test that all NPMI values are greater than the threshold:
    assert all(item[1] > threshold for npmi_list in sorted_npmi.values() for item in npmi_list), "All NPMI values should be greater than the threshold."


def test_get_sorted_npmi_values_with_top_n(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    top_n = 1
    sorted_npmi = npmi_estimator_complex.get_sorted_npmi_values(top_n=top_n)
    # Since unigrams is empty, we can only test bigram and trigram NPMI values:
    bigram_npmi = sorted_npmi[2]
    bigram_npmi = [(item[0], round(item[1], 2)) for item in bigram_npmi]
    assert bigram_npmi == [('test example', round(2.1918065485787688, 2))]
    trigram_npmi = sorted_npmi[3]
    assert trigram_npmi == [('test example sample', 0)]
    assert all(len(npmi_list) <= top_n for n, npmi_list in sorted_npmi.items() if n > 1), "Only the top N NPMI values should be returned for each n."


def test_get_sorted_npmi_values_with_reverse(npmi_estimator_complex: NpmiEstimator) -> None:
    npmi_estimator_complex.estimate_within_corpus_npmi()
    sorted_npmi = npmi_estimator_complex.get_sorted_npmi_values(reverse=False)
    # Since unigrams is empty, we can only test bigram and trigram NPMI values:
    bigram_npmi = sorted_npmi[2]
    bigram_npmi = [(item[0], round(item[1], 2)) for item in bigram_npmi]
    assert bigram_npmi == [('test sample', round(0.17987631177945834, 2)), ('example test', round(0.7268330278608419, 2)), ('test example', round(2.1918065485787688, 2))]
    trigram_npmi = sorted_npmi[3]
    assert trigram_npmi == [('test example sample', 0)]
